# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.tests.utils import BuilderTestCase


class TestBuilderGroup(BuilderTestCase):
    def test_nested_groups_diagram(self):
        diagram = self.build('nested_groups.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'Z': (0, 2)})

    def test_nested_groups_and_edges_diagram(self):
        diagram = self.build('nested_groups_and_edges.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_empty_group_diagram(self):
        diagram = self.build('empty_group.diag')
        self.assertNodeXY(diagram, {'Z': (0, 0)})

    def test_empty_nested_group_diagram(self):
        diagram = self.build('empty_nested_group.diag')
        self.assertNodeXY(diagram, {'Z': (0, 0)})

    def test_empty_group_declaration_diagram(self):
        diagram = self.build('empty_group_declaration.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'Z': (0, 1)})

    def test_simple_group_diagram(self):
        diagram = self.build('simple_group.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'Z': (0, 2)})

    def test_group_declare_as_node_attribute_diagram(self):
        diagram = self.build('group_declare_as_node_attribute.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'E': (2, 2), 'Z': (0, 3)})

    def test_group_attribute(self):
        diagram = self.build('group_attribute.diag')
        groups = list(diagram.traverse_groups())
        self.assertEqual(1, len(groups))
        self.assertEqual((255, 0, 0), groups[0].color)
        self.assertEqual('line', groups[0].shape)

    def test_merge_groups_diagram(self):
        diagram = self.build('merge_groups.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (0, 1), 'D': (1, 1),
                                    'Z': (0, 2)})

    def test_node_attribute_and_group_diagram(self):
        diagram = self.build('node_attribute_and_group.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})
        self.assertNodeLabel(diagram, {'A': 'foo', 'B': 'bar',
                                       'C': 'baz', 'Z': 'Z'})
        self.assertNodeColor(diagram, {'A': (255, 0, 0), 'B': '#888888',
                                       'C': (0, 0, 255), 'Z': (255, 255, 255)})

    def test_group_sibling_diagram(self):
        diagram = self.build('group_sibling.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 2), 'D': (2, 0),
                                    'E': (2, 1), 'F': (2, 2),
                                    'Z': (0, 3)})

    def test_group_order_diagram(self):
        diagram = self.build('group_order.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'Z': (0, 2)})

    def test_group_order2_diagram(self):
        diagram = self.build('group_order2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (2, 1),
                                    'E': (1, 2), 'F': (2, 2),
                                    'Z': (0, 3)})

    def test_group_order3_diagram(self):
        diagram = self.build('group_order3.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'E': (1, 2), 'Z': (0, 3)})

    def test_group_children_height_diagram(self):
        diagram = self.build('group_children_height.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (2, 0), 'F': (2, 2),
                                    'Z': (0, 3)})

    def test_group_children_order_diagram(self):
        diagram = self.build('group_children_order.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (2, 0), 'F': (2, 1),
                                    'G': (2, 2), 'Z': (0, 3)})

    def test_group_children_order2_diagram(self):
        diagram = self.build('group_children_order2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (2, 1), 'F': (2, 0),
                                    'G': (2, 2), 'Z': (0, 3)})

    def test_group_children_order3_diagram(self):
        diagram = self.build('group_children_order3.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (2, 0), 'F': (2, 1),
                                    'G': (2, 2), 'Q': (0, 3),
                                    'Z': (0, 4)})

    def test_group_children_order4_diagram(self):
        diagram = self.build('group_children_order4.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (2, 0), 'Z': (0, 3)})

    def test_node_in_group_follows_outer_node_diagram(self):
        diagram = self.build('node_in_group_follows_outer_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_group_id_and_node_id_are_not_conflicted_diagram(self):
        diagram = self.build('group_id_and_node_id_are_not_conflicted.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (0, 1), 'D': (1, 1),
                                    'Z': (0, 2)})

    def test_outer_node_follows_node_in_group_diagram(self):
        diagram = self.build('outer_node_follows_node_in_group.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_large_group_and_node_diagram(self):
        diagram = self.build('large_group_and_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (1, 3), 'F': (2, 0),
                                    'Z': (0, 4)})

    def test_large_group_and_node2_diagram(self):
        diagram = self.build('large_group_and_node2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (3, 0),
                                    'E': (4, 0), 'F': (5, 0),
                                    'Z': (0, 1)})

    def test_large_group_and_two_nodes_diagram(self):
        diagram = self.build('large_group_and_two_nodes.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (1, 2),
                                    'E': (1, 3), 'F': (2, 0),
                                    'G': (2, 1), 'Z': (0, 4)})

    def test_group_height_diagram(self):
        diagram = self.build('group_height.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'E': (1, 2), 'Z': (0, 3)})

    def test_multiple_groups_diagram(self):
        diagram = self.build('multiple_groups.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (0, 2), 'D': (0, 3),
                                    'E': (1, 0), 'F': (1, 1),
                                    'G': (1, 2), 'H': (2, 0),
                                    'I': (2, 1), 'J': (3, 0),
                                    'Z': (0, 4)})

    def test_multiple_nested_groups_diagram(self):
        diagram = self.build('multiple_nested_groups.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'Z': (0, 2)})

    def test_group_works_node_decorator_diagram(self):
        diagram = self.build('group_works_node_decorator.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (3, 0), 'D': (2, 0),
                                    'E': (1, 1), 'Z': (0, 2)})

    def test_nested_groups_work_node_decorator_diagram(self):
        diagram = self.build('nested_groups_work_node_decorator.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'Z': (0, 2)})

    def test_reversed_multiple_groups_diagram(self):
        diagram = self.build('reverse_multiple_groups.diag')
        self.assertNodeXY(diagram, {'A': (3, 0), 'B': (3, 1),
                                    'C': (3, 2), 'D': (3, 3),
                                    'E': (2, 0), 'F': (2, 1),
                                    'G': (2, 2), 'H': (1, 0),
                                    'I': (1, 1), 'J': (0, 0),
                                    'Z': (0, 4)})

    def test_group_and_skipped_edge_diagram(self):
        diagram = self.build('group_and_skipped_edge.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (3, 0),
                                    'E': (1, 1), 'Z': (0, 2)})

    def test_group_orientation_diagram(self):
        diagram = self.build('group_orientation.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (2, 1),
                                    'Z': (0, 2)})

    def test_nested_group_orientation_diagram(self):
        diagram = self.build('nested_group_orientation.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (1, 0), 'Z': (0, 2)})



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import unicodedata
from functools import wraps
from blockdiag.utils import Size
from blockdiag.utils.compat import u


def is_zenkaku(char):
    """Detect given character is Japanese ZENKAKU character"""
    char_width = unicodedata.east_asian_width(char)
    return char_width in u("WFA")


def zenkaku_len(string):
    """Count Japanese ZENKAKU characters from string"""
    return len([x for x in string if is_zenkaku(x)])


def hankaku_len(string):
    """Count non Japanese ZENKAKU characters from string"""
    return len([x for x in string if not is_zenkaku(x)])


def string_width(string):
    """Measure rendering width of string.
       Count ZENKAKU-character as 2-point and non ZENKAKU-character as 1-point
    """
    widthmap = {'Na': 1, 'N': 1, 'H': 1, 'W': 2, 'F': 2, 'A': 2}
    return sum(widthmap[unicodedata.east_asian_width(c)] for c in string)


def textsize(string, font):
    """Measure rendering size (width and height) of line.
       Returned size will not be exactly as rendered text size,
       Because this method does not use fonts to measure size.
    """
    width = (zenkaku_len(string) * font.size +
             hankaku_len(string) * font.size * 0.55)

    return Size(int(math.ceil(width)), font.size)


def memoize(fn):
    fn.cache = {}

    @wraps(fn)
    def func(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in fn.cache:
            fn.cache[key] = fn(*args, **kwargs)

        return fn.cache[key]

    return func



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import os
import copy
from collections import namedtuple
from blockdiag.utils.config import ConfigParser
from blockdiag.utils.logging import warning


def parse_fontpath(path):
    if path is None:
        return (None, None)

    match = re.search('^(.*):(\d)$', path)
    if match:
        return (match.group(1), int(match.group(2)))
    else:
        return (path, None)


class FontInfo(object):
    def __init__(self, family, path, size):
        self.path = path
        self.size = int(size)

        family = self._parse(family)
        self.name = family[0]
        self.generic_family = family[1]
        self.weight = family[2]
        self.style = family[3]

    def __repr__(self):
        return ("<FontInfo familyname=%r size=%r>" %
                (self.familyname, self.size))

    @property
    def familyname(self):
        if self.name:
            name = self.name + "-"
        else:
            name = ''

        if self.generic_family == 'sans-serif':
            generic_family = 'sansserif'
        else:
            generic_family = self.generic_family

        if self.weight == 'bold':
            return "%s%s-%s" % (name, generic_family, self.weight)
        else:
            return "%s%s-%s" % (name, generic_family, self.style)

    def _parse(self, familyname):
        pattern = '^(?:(.*)-)?' + \
                  '(serif|sansserif|monospace|fantasy|cursive)' + \
                  '(?:-(normal|bold|italic|oblique))?$'

        match = re.search(pattern, familyname or '')
        if match is None:
            msg = 'Unknown font family: %s' % familyname
            raise AttributeError(msg)

        name = match.group(1) or ''
        generic_family = match.group(2)
        style = match.group(3) or ''

        if generic_family == 'sansserif':
            generic_family = 'sans-serif'

        if style == 'bold':
            weight = 'bold'
            style = 'normal'
        elif style in ('italic', 'oblique'):
            weight = 'normal'
            style = style
        else:
            weight = 'normal'
            style = 'normal'

        return [name, generic_family, weight, style]

    def duplicate(self):
        return copy.copy(self)


class FontMap(object):
    BASE_FONTSIZE = 11
    fontsize = BASE_FONTSIZE
    default_fontfamily = 'sansserif'

    def __init__(self, filename=None):
        self.fonts = {}
        self.aliases = {}

        if filename:
            self._parse_config(filename)
        self.set_default_font(None)

    def set_default_fontfamily(self, fontfamily):
        self.default_fontfamily = fontfamily
        self.set_default_font(None)

    def _parse_config(self, conffile):
        config = ConfigParser()

        if hasattr(conffile, 'read'):
            config.readfp(conffile)
        elif os.path.isfile(conffile):
            config.read(conffile)
        else:
            msg = "fontmap file is not found: %s" % conffile
            raise RuntimeError(msg)

        if config.has_section('fontmap'):
            for name, path in config.items('fontmap'):
                self.append_font(name, path)

        if config.has_section('fontalias'):
            for name, family in config.items('fontalias'):
                self.aliases[name] = family

    def set_default_font(self, path):
        if path is None and self.find() is not None:
            return

        self.append_font(self.default_fontfamily, path)

    def append_font(self, fontfamily, path):
        _path, _ = parse_fontpath(path)
        if path is None or os.path.isfile(_path):
            font = FontInfo(fontfamily, path, self.fontsize)
            self.fonts[font.familyname] = font
        else:
            warning('fontfile `%s` is not found: %s', fontfamily, path)

    def _regulate_familyname(self, name):
        return FontInfo(name, None, self.BASE_FONTSIZE).familyname.lower()

    def find(self, element=None):
        fontfamily = getattr(element, 'fontfamily', None) or \
            self.default_fontfamily
        fontfamily = self.aliases.get(fontfamily, fontfamily)
        fontsize = getattr(element, 'fontsize', None) or self.fontsize

        name = self._regulate_familyname(fontfamily)
        if name in self.fonts:
            font = self.fonts[name].duplicate()
            font.size = fontsize
        elif element is not None:
            warning("Unknown fontfamily: %s", fontfamily)
            elem = namedtuple('Font', 'fontsize')(fontsize)
            font = self.find(elem)
        else:
            font = None

        return font



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from blockdiag.utils import Size, unquote

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestUtils(unittest.TestCase):
    def test_size_resize(self):
        size = Size(10, 20)

        resized = size.resize(width=50, height=50)
        self.assertEqual((50, 50), resized)

        resized = size.resize(width=50)
        self.assertEqual((50, 100), resized)

        resized = size.resize(height=50)
        self.assertEqual((25, 50), resized)

        resized = size.resize(scale=50)
        self.assertEqual((5, 10), resized)

        resized = size.resize(width=50, scale=50)
        self.assertEqual((25, 50), resized)

        resized = size.resize(height=50, scale=50)
        self.assertEqual((12.5, 25), resized)

        resized = size.resize(width=50, height=50, scale=50)
        self.assertEqual((25, 25), resized)

    def test_size_to_integer_point(self):
        size = Size(1.5, 2.5)

        self.assertEqual((1, 2), size.to_integer_point())

    def test_unquote(self):
        self.assertEqual('test', unquote('"test"'))
        self.assertEqual('test', unquote("'test'"))
        self.assertEqual("'half quoted", unquote("'half quoted"))
        self.assertEqual('"half quoted', unquote('"half quoted'))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pkg_resources
from blockdiag.utils.logging import warning

drawers = {}


def init_imagedrawers(debug=False):
    for drawer in pkg_resources.iter_entry_points('blockdiag_imagedrawers'):
        try:
            module = drawer.load()
            if hasattr(module, 'setup'):
                module.setup(module)
        except Exception as exc:
            if debug:
                warning('Failed to load %s: %r' % (drawer.module_name, exc))


def install_imagedrawer(ext, drawer):
    drawers[ext] = drawer


def create(_format, filename, **kwargs):
    if len(drawers) == 0:
        init_imagedrawers(debug=kwargs.get('debug'))

    _format = _format.lower()
    if _format in drawers:
        drawer = drawers[_format](filename, **kwargs)
    else:
        msg = 'failed to load %s image driver' % _format
        raise RuntimeError(msg)

    if 'linejump' in kwargs.get('filters', []):
        from blockdiag.imagedraw.filters.linejump import LineJumpDrawFilter
        jumpsize = kwargs.get('jumpsize', 0)
        drawer = LineJumpDrawFilter(drawer, jumpsize)

    return drawer



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
import io
import os
import re
from PIL import Image
from tempfile import NamedTemporaryFile
from blockdiag.utils import urlutil
from blockdiag.utils.compat import u
from blockdiag.utils.logging import warning

urlopen_cache = {}


def urlopen(url, *args, **kwargs):
    """ auto caching urlopen() (using tempfile) """
    from blockdiag.utils.compat import urlopen as orig_urlopen

    if url not in urlopen_cache:
        with NamedTemporaryFile(delete=False) as tmpfile:
            tmpfile.write(orig_urlopen(url, *args, **kwargs).read())
            tmpfile.flush()
            urlopen_cache[url] = tmpfile.name

    return io.open(urlopen_cache[url], 'rb')


def get_image_size(image):
    if isinstance(image, Image.Image):
        return image.size
    else:
        stream = None
        try:
            stream = open(image)
            return stream.size
        finally:
            if stream and hasattr(stream, 'close'):
                stream.close()


def calc_image_size(size, bounded):
    if bounded[0] < size[0] or bounded[1] < size[1]:
        if (size[0] * 1.0 // bounded[0]) < (size[1] * 1.0 // bounded[1]):
            size = (size[0] * bounded[1] // size[1], bounded[1])
        else:
            size = (bounded[0], size[1] * bounded[0] // size[0])

    return size


def color_to_rgb(color):
    import webcolors
    if color == 'none' or isinstance(color, (list, tuple)):
        rgb = color
    elif re.match('#', color):
        rgb = webcolors.hex_to_rgb(color)
    else:
        rgb = webcolors.name_to_rgb(color)

    return rgb


def wand_open(url, stream):
    try:
        import wand.image
    except:
        warning("unknown image type: %s", url)
        raise IOError

    try:
        png_image = io.BytesIO()
        with wand.image.Image(file=stream) as img:
            img.format = 'PNG'
            img.save(file=png_image)
            png_image.seek(0)
            return png_image
    except Exception as exc:
        warning("Fail to convert %s to PNG: %r", url, exc)
        raise IOError


def pillow_open(url, stream):
    try:
        if isinstance(url, Image.Image):
            return url
        else:
            return Image.open(stream)
    except IOError:
        stream.seek(0)
        png_stream = wand_open(url, stream)

        return Image.open(png_stream)


def open(url, mode='Pillow'):
    if hasattr(url, 'read') or isinstance(url, Image.Image):
        stream = url
    elif not urlutil.isurl(url):
        stream = io.open(url, 'rb')
    else:
        try:
            # wrap BytesIO for rewind stream
            stream = io.BytesIO(urlopen(url).read())
        except:
            warning(u("Could not retrieve: %s"), url)
            raise IOError

    image = pillow_open(url, stream)
    if mode.lower() == 'pillow':
        # stream will be closed by GC
        return image
    else:  # mode == 'png'
        try:
            png_image = io.BytesIO()
            image.save(png_image, 'PNG')
            if hasattr(stream, 'close'):  # close() is implemented on Pillow
                stream.close()
        except:
            warning(u("Could not convert image: %s"), url)
            raise IOError

        png_image.seek(0)
        return png_image


def cleanup():
    for url in list(urlopen_cache.keys()):
        path = urlopen_cache.pop(url)
        try:
            os.remove(path)
        except:
            pass


def setup(app):
    app.register_cleanup_handler(cleanup)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import cycle
from itertools import islice


def istep(seq, step=2):
    iterable = iter(seq)
    while True:
        item = list(islice(iterable, step))
        if len(item) < step:
            break
        yield item


def stepslice(iterable, steps):
    try:
        iterable = iter(iterable)
        step = cycle(steps)

        while True:
            # skip (1)
            n = next(step)
            if n == 0:
                pass
            elif n == 1:
                o = next(iterable)
                yield o
                yield o
            else:
                yield next(iterable)
                for _ in range(n - 2):
                    next(iterable)
                yield next(iterable)

            # skip (2)
            for _ in range(next(step)):
                next(iterable)
    except StopIteration:
        return



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
from blockdiag.builder import SeparateDiagramBuilder
from blockdiag.elements import DiagramNode
from blockdiag.tests.utils import BuilderTestCase


class TestBuilderSeparated(BuilderTestCase):
    def _build(self, tree):
        return SeparateDiagramBuilder.build(tree)

    def test_separate1_diagram(self):
        diagram = self.build('separate1.diag')

        assert_pos = {0: {'B': (0, 0), 'C': (1, 0), 'D': (4, 0),
                          'E': (2, 0), 'F': (3, 0)},
                      1: {'A': (0, 0), 'B': (1, 0), 'D': (3, 0)},
                      2: {'A': (0, 0), 'Z': (0, 1)}}

        for i, diagram in enumerate(diagram):
            for node in diagram.traverse_nodes():
                if isinstance(node, DiagramNode):
                    print(node)
                    self.assertEqual(assert_pos[i][node.id], node.xy)

    def test_separate2_diagram(self):
        diagram = self.build('separate2.diag')

        assert_pos = {0: {'A': (0, 0), 'C': (1, 0), 'D': (2, 0),
                          'E': (0, 2), 'G': (3, 0), 'H': (3, 1)},
                      1: {'A': (0, 0), 'B': (1, 0), 'E': (2, 0),
                          'F': (4, 2), 'G': (4, 0), 'H': (4, 1)},
                      2: {'A': (0, 0), 'F': (2, 2), 'G': (2, 0),
                          'H': (2, 1), 'Z': (0, 3)}}

        for i, diagram in enumerate(diagram):
            for node in diagram.traverse_nodes():
                if isinstance(node, DiagramNode):
                    print(node)
                    self.assertEqual(assert_pos[i][node.id], node.xy)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import division
from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import XY, Box
from blockdiag.imagedraw.simplesvg import pathdata


class Database(NodeShape):
    def __init__(self, node, metrics=None):
        super(Database, self).__init__(node, metrics)

        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize
        self.textbox = Box(m.topleft.x, m.topleft.y + r * 3 // 2,
                           m.bottomright.x, m.bottomright.y - r // 2)

    def render_shape(self, drawer, _, **kwargs):
        # draw background
        self.render_shape_background(drawer, **kwargs)

        # draw background image
        if self.node.background:
            drawer.image(self.textbox, self.node.background)

    def render_shape_background(self, drawer, **kwargs):
        fill = kwargs.get('fill')

        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize
        box = m.box

        ellipse = Box(box[0], box[3] - r * 2, box[2], box[3])
        if kwargs.get('shadow'):
            ellipse = self.shift_shadow(ellipse)
            if kwargs.get('style') == 'blur':
                drawer.ellipse(ellipse, fill=fill, outline=fill,
                               filter='transp-blur')
            else:
                drawer.ellipse(ellipse, fill=fill, outline=fill)
        else:
            drawer.ellipse(ellipse, fill=self.node.color,
                           outline=self.node.linecolor, style=self.node.style)

        rect = Box(box[0], box[1] + r, box[2], box[3] - r)
        if kwargs.get('shadow'):
            rect = self.shift_shadow(rect)
            if kwargs.get('style') == 'blur':
                drawer.rectangle(rect, fill=fill, outline=fill,
                                 filter='transp-blur')
            else:
                drawer.rectangle(rect, fill=fill, outline=fill)
        else:
            drawer.rectangle(rect, fill=self.node.color,
                             outline=self.node.color)

        ellipse = Box(box[0], box[1], box[2], box[1] + r * 2)
        if kwargs.get('shadow'):
            ellipse = self.shift_shadow(ellipse)
            if kwargs.get('style') == 'blur':
                drawer.ellipse(ellipse, fill=fill, outline=fill,
                               filter='transp-blur')
            else:
                drawer.ellipse(ellipse, fill=fill, outline=fill)
        else:
            drawer.ellipse(ellipse, fill=self.node.color,
                           outline=self.node.linecolor, style=self.node.style)

        # line both side
        lines = [(XY(box[0], box[1] + r), XY(box[0], box[3] - r)),
                 (XY(box[2], box[1] + r), XY(box[2], box[3] - r))]
        for line in lines:
            if not kwargs.get('shadow'):
                drawer.line(line, fill=self.node.linecolor,
                            style=self.node.style)

    def render_vector_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize
        width = self.metrics.node_width

        box = m.box
        if kwargs.get('shadow'):
            box = self.shift_shadow(box)

        path = pathdata(box[0], box[1] + r)
        path.ellarc(width // 2, r, 0, 0, 1, box[2], box[1] + r)
        path.line(box[2], box[3] - r)
        path.ellarc(width // 2, r, 0, 0, 1, box[0], box[3] - r)
        path.line(box[0], box[1] + r)

        # draw outline
        if kwargs.get('shadow'):
            if kwargs.get('style') == 'blur':
                drawer.path(path, fill=fill, outline=fill,
                            filter='transp-blur')
            else:
                drawer.path(path, fill=fill, outline=fill)
        elif self.node.background:
            drawer.path(path, fill=self.node.color,
                        outline=self.node.color)
            drawer.image(self.textbox, self.node.background)
            drawer.path(path, fill="none",
                        outline=self.node.linecolor, style=self.node.style)
        else:
            drawer.path(path, fill=self.node.color,
                        outline=self.node.linecolor, style=self.node.style)

        # draw cap of cylinder
        if not kwargs.get('shadow'):
            path = pathdata(box[2], box[1] + r)
            path.ellarc(width // 2, r, 0, 0, 1, box[0], box[1] + r)
            drawer.path(path, fill=self.node.color,
                        outline=self.node.linecolor, style=self.node.style)


def setup(self):
    install_renderer('flowchart.database', Database)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.tests.utils import BuilderTestCase


class TestBuilder(BuilderTestCase):
    def test_diagram_attributes(self):
        diagram = self.build('diagram_attributes.diag')

        self.assertEqual(160, diagram.node_width)
        self.assertEqual(160, diagram.node_height)
        self.assertEqual(32, diagram.span_width)
        self.assertEqual(32, diagram.span_height)
        self.assertEqual((128, 128, 128), diagram.linecolor)       # gray
        self.assertEqual('diamond', diagram.nodes[0].shape)
        self.assertEqual('dotted', diagram.nodes[0].style)
        self.assertEqual((255, 0, 0), diagram.nodes[0].color)      # red
        self.assertEqual((0, 128, 0), diagram.nodes[0].textcolor)  # green
        self.assertEqual(16, diagram.nodes[0].fontsize)
        self.assertEqual((0, 0, 255), diagram.nodes[1].color)      # blue
        self.assertEqual((0, 128, 0), diagram.nodes[1].textcolor)  # green
        self.assertEqual(16, diagram.nodes[1].fontsize)

        self.assertEqual((128, 128, 128), diagram.edges[0].color)  # gray
        self.assertEqual((0, 128, 0), diagram.edges[0].textcolor)  # green
        self.assertEqual(16, diagram.edges[0].fontsize)

    def test_diagram_attributes_order_diagram(self):
        diagram = self.build('diagram_attributes_order.diag')
        self.assertNodeColor(diagram, {'A': (255, 0, 0), 'B': (255, 0, 0)})
        self.assertNodeLineColor(diagram, {'A': (255, 0, 0), 'B': (255, 0, 0)})

    def test_circular_ref_to_root_diagram(self):
        diagram = self.build('circular_ref_to_root.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'Z': (0, 2)})

    def test_circular_ref_diagram(self):
        diagram = self.build('circular_ref.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (2, 1),
                                    'Z': (0, 2)})

    def test_circular_ref2_diagram(self):
        diagram = self.build('circular_ref2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (3, 0),
                                    'E': (3, 1), 'F': (4, 0),
                                    'Z': (0, 2)})

    def test_circular_ref_and_parent_node_diagram(self):
        diagram = self.build('circular_ref_and_parent_node.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'D': (2, 1),
                                    'Z': (0, 2)})

    def test_labeled_circular_ref_diagram(self):
        diagram = self.build('labeled_circular_ref.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (2, 0),
                                    'C': (1, 0), 'Z': (0, 1)})

    def test_twin_forked_diagram(self):
        diagram = self.build('twin_forked.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 2), 'D': (2, 0),
                                    'E': (3, 0), 'F': (3, 1),
                                    'G': (4, 1), 'Z': (0, 3)})

    def test_skipped_edge_diagram(self):
        diagram = self.build('skipped_edge.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_circular_skipped_edge_diagram(self):
        diagram = self.build('circular_skipped_edge.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'Z': (0, 1)})

    def test_triple_branched_diagram(self):
        diagram = self.build('triple_branched.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (0, 2), 'D': (1, 0),
                                    'Z': (0, 3)})

    def test_twin_circular_ref_to_root_diagram(self):
        diagram = self.build('twin_circular_ref_to_root.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (1, 1), 'Z': (0, 2)})

    def test_twin_circular_ref_diagram(self):
        diagram = self.build('twin_circular_ref.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (1, 1),
                                    'Z': (0, 2)})

    def test_skipped_circular_diagram(self):
        diagram = self.build('skipped_circular.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 1),
                                    'C': (2, 0), 'Z': (0, 2)})

    def test_skipped_twin_circular_diagram(self):
        diagram = self.build('skipped_twin_circular.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 1), 'D': (2, 2),
                                    'E': (3, 0), 'Z': (0, 3)})

    def test_nested_skipped_circular_diagram(self):
        diagram = self.build('nested_skipped_circular.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 1), 'D': (3, 2),
                                    'E': (4, 1), 'F': (5, 0),
                                    'G': (6, 0), 'Z': (0, 3)})

    def test_self_ref_diagram(self):
        diagram = self.build('self_ref.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'Z': (0, 1)})

    def test_diagram_orientation_diagram(self):
        diagram = self.build('diagram_orientation.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (0, 2), 'D': (1, 2),
                                    'Z': (2, 0)})

    def test_nested_group_orientation2_diagram(self):
        diagram = self.build('nested_group_orientation2.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (0, 1),
                                    'C': (0, 2), 'D': (1, 2),
                                    'E': (2, 2), 'F': (2, 3),
                                    'Z': (3, 0)})

    def test_slided_children_diagram(self):
        diagram = self.build('slided_children.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (1, 3),
                                    'E': (2, 3), 'F': (3, 2),
                                    'G': (2, 1), 'H': (4, 1)})

    def test_non_rhombus_relation_height_diagram(self):
        diagram = self.build('non_rhombus_relation_height.diag')
        self.assertNodeXY(diagram, {'A': (0, 0), 'B': (1, 0),
                                    'C': (2, 0), 'D': (0, 1),
                                    'E': (0, 2), 'F': (1, 2),
                                    'G': (1, 3), 'H': (2, 3),
                                    'I': (2, 4), 'J': (1, 5),
                                    'K': (2, 5), 'Z': (0, 6)})

    def test_define_class_diagram(self):
        diagram = self.build('define_class.diag')
        self.assertNodeColor(diagram, {'A': (255, 0, 0),
                                       'B': (255, 255, 255),
                                       'C': (255, 255, 255)})
        self.assertNodeStyle(diagram, {'A': 'dashed', 'B': None, 'C': None})
        self.assertEdgeColor(diagram, {('A', 'B'): (255, 0, 0),
                                       ('B', 'C'): (0, 0, 0)})
        self.assertEdgeStyle(diagram, {('A', 'B'): 'dashed',
                                       ('B', 'C'): None})



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
import codecs

if sys.version_info[0] == 2:
    string_types = (str, unicode)  # NOQA: pyflakes complains to unicode in py3
    from urllib import urlopen  # NOQA: exporting for common interface
else:
    string_types = (str,)
    from urllib.request import urlopen  # NOQA: exporting for common interface


def u(string):
    if sys.version_info[0] == 2:
        return unicode(string, "unicode_escape")  # NOQA: pyflakes complains to unicode in py3
    else:
        return string


# replace codecs.getreader
if sys.version_info[0] == 3:
    getreader = codecs.getreader

    def py3_getreader(encoding):
        return lambda stream, *args: getreader(encoding)(stream.buffer, *args)

    codecs.getreader = py3_getreader


def cmp_to_key(mycmp):
    """Convert a cmp= function into a key= function"""
    class K(object):
        __slots__ = ['obj']

        def __init__(self, obj, *args):
            self.obj = obj

        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0

        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0

        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0

        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0

        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0

        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0

        def __hash__(self):
            raise TypeError('hash not implemented')

    return K



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import re
import sys
from nose.tools import nottest
import blockdiag
import blockdiag.command
from blockdiag.tests.utils import capture_stderr, TemporaryDirectory
from blockdiag.tests.utils import supported_pil, supported_pdf

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


TESTDIR = os.path.dirname(__file__)
FONTPATH = os.path.join(TESTDIR, 'VLGothic', 'VL-Gothic-Regular.ttf')


def get_fontpath(testdir):
    return os.path.join(testdir, 'VLGothic', 'VL-Gothic-Regular.ttf')


def get_diagram_files(testdir):
    diagramsdir = os.path.join(testdir, 'diagrams')

    skipped = ['README', 'debian-logo-256color-palettealpha.png',
               'errors', 'white.gif']
    for file in os.listdir(diagramsdir):
        if file in skipped:
            pass
        else:
            yield os.path.join(diagramsdir, file)


def test_generate():
    mainfunc = blockdiag.command.main
    basepath = os.path.dirname(__file__)
    files = get_diagram_files(basepath)
    options = []

    for testcase in testcase_generator(basepath, mainfunc, files, options):
        yield testcase


def test_generate_with_separate():
    mainfunc = blockdiag.command.main
    basepath = os.path.dirname(__file__)
    files = get_diagram_files(basepath)
    filtered = (f for f in files if re.search('separate', f))
    options = ['--separate']

    for testcase in testcase_generator(basepath, mainfunc, filtered, options):
        yield testcase


@nottest
def testcase_generator(basepath, mainfunc, files, options):
    fontpath = get_fontpath(basepath)
    options = options + ['-f', fontpath]

    for source in files:
        yield generate, mainfunc, 'svg', source, options

        if not supported_pil():
            yield unittest.skip("Pillow is not available")(generate)
            yield unittest.skip("Pillow is not available")(generate)
        elif os.environ.get('ALL_TESTS') is None:
            message = "Skipped by default. To enable it, specify $ALL_TESTS=1"
            yield unittest.skip(message)(generate)
            yield unittest.skip(message)(generate)
        else:
            yield generate, mainfunc, 'png', source, options
            yield generate, mainfunc, 'png', source, options + ['--antialias']

        if not supported_pdf():
            yield unittest.skip("reportlab is not available")(generate)
        elif os.environ.get('ALL_TESTS') is None:
            message = "Skipped by default. To enable it, specify $ALL_TESTS=1"
            yield unittest.skip(message)(generate)
        else:
            yield generate, mainfunc, 'pdf', source, options


@capture_stderr
def generate(mainfunc, filetype, source, options):
    try:
        tmpdir = TemporaryDirectory()
        fd, tmpfile = tmpdir.mkstemp()
        os.close(fd)

        mainfunc(['--debug', '-T', filetype, '-o', tmpfile, source] +
                 list(options))
    finally:
        tmpdir.clean()


def not_exist_font_config_option_test():
    args = ['-f', '/font_is_not_exist', '-f', FONTPATH, 'input.diag']
    options = blockdiag.command.BlockdiagOptions(blockdiag).parse(args)

    from blockdiag.utils.bootstrap import detectfont
    detectfont(options)


def stdin_test():
    testdir = os.path.dirname(__file__)
    diagpath = os.path.join(testdir, 'diagrams', 'single_edge.diag')

    try:
        stdin = sys.stdin
        sys.stdin = open(diagpath, 'r')

        tmpdir = TemporaryDirectory()
        fd, tmpfile = tmpdir.mkstemp()
        os.close(fd)

        args = ['-T', 'SVG', '-o', tmpfile, '-']
        ret = blockdiag.command.main(args)
        assert ret == 0
    finally:
        sys.stdin = stdin
        tmpdir.clean()


@capture_stderr
def ghostscript_not_found_test():
    testdir = os.path.dirname(__file__)
    diagpath = os.path.join(testdir, 'diagrams', 'background_url_image.diag')

    try:
        old_path = os.environ['PATH']
        os.environ['PATH'] = ''
        tmpdir = TemporaryDirectory()
        fd, tmpfile = tmpdir.mkstemp()
        os.close(fd)

        args = ['-T', 'SVG', '-o', tmpfile, diagpath]
        ret = blockdiag.command.main(args)
        assert 'Could not convert image:' in sys.stderr.getvalue()
        assert ret == 0
    finally:
        tmpdir.clean()
        os.environ['PATH'] = old_path


@capture_stderr
def svg_includes_source_code_tag_test():
    from xml.etree import ElementTree

    testdir = os.path.dirname(__file__)
    diagpath = os.path.join(testdir, 'diagrams', 'single_edge.diag')

    try:
        tmpdir = TemporaryDirectory()
        fd, tmpfile = tmpdir.mkstemp()
        os.close(fd)

        args = ['-T', 'SVG', '-o', tmpfile, diagpath]
        blockdiag.command.main(args)

        # compare embeded source code
        source_code = open(diagpath).read()
        tree = ElementTree.parse(tmpfile)
        desc = tree.find('{http://www.w3.org/2000/svg}desc')

        # strip spaces
        source_code = re.sub('\s+', ' ', source_code)
        embeded = re.sub('\s+', ' ', desc.text)
        assert source_code == embeded
    finally:
        tmpdir.clean()



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import sys
import blockdiag
import blockdiag.builder
import blockdiag.drawer
import blockdiag.parser
from blockdiag.utils.bootstrap import Application, Options


class BlockdiagOptions(Options):
    def build_parser(self):
        super(BlockdiagOptions, self).build_parser()
        self.parser.add_option(
            '-s', '--separate', action='store_true',
            help='Separate diagram images for each group (SVG only)'
        )


class BlockdiagApp(Application):
    module = blockdiag

    def parse_options(self, args):
        self.options = BlockdiagOptions(self.module).parse(args)

    def build_diagram(self, tree):
        if not self.options.separate:
            return super(BlockdiagApp, self).build_diagram(tree)
        else:
            DiagramBuilder = self.module.builder.SeparateDiagramBuilder
            DiagramDraw = self.module.drawer.DiagramDraw

            basename = re.sub('.svg$', '', self.options.output)
            for i, group in enumerate(DiagramBuilder.build(tree)):
                outfile = '%s_%d.svg' % (basename, i + 1)
                draw = DiagramDraw(self.options.type, group, outfile,
                                   fontmap=self.fontmap,
                                   antialias=self.options.antialias,
                                   nodoctype=self.options.nodoctype,
                                   transparency=self.options.transparency)
                draw.draw()
                draw.save()

            return 0


def main(args=sys.argv[1:]):
    return BlockdiagApp().run(args)



# -*- coding: utf-8 -*-
import sys
from setuptools import setup, find_packages

sys.path.insert(0, 'src')
import blockdiag

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.6",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Topic :: Software Development",
    "Topic :: Software Development :: Documentation",
    "Topic :: Text Processing :: Markup",
]

requires = ['setuptools',
            'funcparserlib',
            'Pillow']
pdf_requires = ['reportlab']
test_requires = ['nose',
                 'mock',
                 'pep8>=1.3',
                 'reportlab',
                 'docutils']


# only for Python2.6
if (2, 6) < sys.version_info < (2, 7):
    requires.append('OrderedDict')
    pdf_requires[0] = 'reportlab < 3.0'
    test_requires.append('unittest2')

# webcolors
if (3, 2) < sys.version_info < (3, 3):
    requires.append('webcolors < 1.5')  # webcolors-1.5 does not support py32
else:
    requires.append('webcolors')


setup(
    name='blockdiag',
    version=blockdiag.__version__,
    description='blockdiag generates block-diagram image from text',
    long_description=open("README.rst").read(),
    classifiers=classifiers,
    keywords=['diagram', 'generator'],
    author='Takeshi Komiya',
    author_email='i.tkomiya@gmail.com',
    url='http://blockdiag.com/',
    download_url='http://pypi.python.org/pypi/blockdiag',
    license='Apache License 2.0',
    py_modules=['blockdiag_sphinxhelper'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['buildout.cfg']},
    include_package_data=True,
    install_requires=requires,
    extras_require=dict(
        testing=test_requires,
        pdf=pdf_requires,
        rst=[
            'docutils',
        ],
    ),
    test_suite='nose.collector',
    tests_require=test_requires,
    entry_points="""
       [console_scripts]
       blockdiag = blockdiag.command:main

       [blockdiag_noderenderer]
       box = blockdiag.noderenderer.box
       square = blockdiag.noderenderer.square
       roundedbox = blockdiag.noderenderer.roundedbox
       diamond = blockdiag.noderenderer.diamond
       minidiamond = blockdiag.noderenderer.minidiamond
       mail = blockdiag.noderenderer.mail
       note = blockdiag.noderenderer.note
       cloud = blockdiag.noderenderer.cloud
       circle = blockdiag.noderenderer.circle
       ellipse = blockdiag.noderenderer.ellipse
       beginpoint = blockdiag.noderenderer.beginpoint
       endpoint = blockdiag.noderenderer.endpoint
       actor = blockdiag.noderenderer.actor
       flowchart.database = blockdiag.noderenderer.flowchart.database
       flowchart.input = blockdiag.noderenderer.flowchart.input
       flowchart.loopin = blockdiag.noderenderer.flowchart.loopin
       flowchart.loopout = blockdiag.noderenderer.flowchart.loopout
       flowchart.terminator = blockdiag.noderenderer.flowchart.terminator
       textbox = blockdiag.noderenderer.textbox
       dots = blockdiag.noderenderer.dots
       none = blockdiag.noderenderer.none

       [blockdiag_plugins]
       attributes = blockdiag.plugins.attributes
       autoclass = blockdiag.plugins.autoclass

       [blockdiag_imagedrawers]
       imagedraw_png = blockdiag.imagedraw.png
       imagedraw_svg = blockdiag.imagedraw.svg
       imagedraw_pdf = blockdiag.imagedraw.pdf
    """,
)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag.noderenderer import install_renderer
from blockdiag.noderenderer.base import NodeShape
from blockdiag.utils import XY


class Note(NodeShape):
    def render_shape(self, drawer, _, **kwargs):
        fill = kwargs.get('fill')

        m = self.metrics.cell(self.node)
        r = self.metrics.cellsize * 2

        tr = m.topright
        note = [m.topleft, XY(tr.x - r, tr.y), XY(tr.x, tr.y + r),
                m.bottomright, m.bottomleft, m.topleft]
        box = self.metrics.cell(self.node).box

        # draw outline
        if kwargs.get('shadow'):
            note = self.shift_shadow(note)
            if kwargs.get('style') == 'blur':
                drawer.polygon(note, fill=fill, outline=fill,
                               filter='transp-blur')
            else:
                drawer.polygon(note, fill=fill, outline=fill)
        elif self.node.background:
            drawer.polygon(note, fill=self.node.color,
                           outline=self.node.color)
            drawer.image(box, self.node.background)
            drawer.polygon(note, fill="none",
                           outline=self.node.linecolor, style=self.node.style)
        else:
            drawer.polygon(note, fill=self.node.color,
                           outline=self.node.linecolor, style=self.node.style)

        # draw folded
        if not kwargs.get('shadow'):
            folded = [XY(tr.x - r, tr.y),
                      XY(tr.x - r, tr.y + r),
                      XY(tr.x, tr.y + r)]
            drawer.line(folded, fill=self.node.linecolor,
                        style=self.node.style)


def setup(self):
    install_renderer('note', Note)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from blockdiag.imagedraw.utils import (
    is_zenkaku, zenkaku_len, hankaku_len,
    string_width, textsize
)
from blockdiag.utils.compat import u

if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest


class TestUtils(unittest.TestCase):
    def test_is_zenkaku(self):
        # A
        self.assertEqual(False, is_zenkaku(u("A")))
        # あ
        self.assertEqual(True, is_zenkaku(u("\u3042")))

    def test_zenkaku_len(self):
        # abc
        self.assertEqual(0, zenkaku_len(u("abc")))
        # あいう
        self.assertEqual(3, zenkaku_len(u("\u3042\u3044\u3046")))
        # あいc
        self.assertEqual(2, zenkaku_len(u("\u3042\u3044c")))

    def test_hankaku_len(self):
        # abc
        self.assertEqual(3, hankaku_len(u("abc")))
        # あいう
        self.assertEqual(0, hankaku_len(u("\u3042\u3044\u3046")))
        # あいc
        self.assertEqual(1, hankaku_len(u("\u3042\u3044c")))

    def test_string_width(self):
        # abc
        self.assertEqual(3, string_width(u("abc")))
        # あいう
        self.assertEqual(6, string_width(u("\u3042\u3044\u3046")))
        # あいc
        self.assertEqual(5, string_width(u("\u3042\u3044c")))

    def test_test_textsize(self):
        from blockdiag.utils.fontmap import FontInfo
        font = FontInfo('serif', None, 11)

        # abc
        self.assertEqual((19, 11), textsize(u("abc"), font))
        # あいう
        self.assertEqual((33, 11), textsize(u("\u3042\u3044\u3046"), font))
        # あいc
        self.assertEqual((29, 11), textsize(u("\u3042\u3044c"), font))

        # abc
        font = FontInfo('serif', None, 24)
        self.assertEqual((40, 24), textsize(u("abc"), font))

        # あいう
        font = FontInfo('serif', None, 18)
        self.assertEqual((54, 18), textsize(u("\u3042\u3044\u3046"), font))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from blockdiag import plugins


class AutoClass(plugins.NodeHandler):
    def on_created(self, node):
        if node.id is None:
            return True

        for name, klass in self.diagram.classes.items():
            pattern = "_%s$" % re.escape(name)

            if re.search(pattern, node.id):
                node.label = re.sub(pattern, '', node.id)
                node.set_attributes(klass.attrs)

        return True


def setup(self, diagram, **kwargs):
    plugins.install_node_handler(AutoClass(diagram, **kwargs))



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import blockdiag.parser
import blockdiag.builder
import blockdiag.drawer

import blockdiag.utils.bootstrap
import blockdiag.utils.compat
import blockdiag.utils.fontmap
import blockdiag.utils.rst.nodes
import blockdiag.utils.rst.directives

__all__ = [
    'core', 'utils'
]

core = blockdiag
utils = blockdiag.utils



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
from blockdiag.utils import Box, XY


class LazyReceiver(object):
    def __init__(self, target):
        self.target = target
        self.calls = []

    def __getattr__(self, name):
        return self.get_lazy_method(name)

    def get_lazy_method(self, name):
        if name in self.target.nosideeffect_methods:
            method = self._find_method(name)
            return functools.partial(method, self.target)
        elif name in self.target.self_generative_methods:
            def _(*args, **kwargs):
                receiver = LazySubReceiver(name, self.target, *args, **kwargs)
                self.calls.append((receiver, args, kwargs))
                return receiver

            return _
        else:
            def _(*args, **kwargs):
                self.calls.append((name, args, kwargs))
                return self

            return _

    def _find_method(self, name):
        if isinstance(name, LazyReceiver):
            return name

        for p in self.target.__class__.__mro__:
            if name in p.__dict__:
                return p.__dict__[name]

        raise AttributeError("%s instance has no attribute '%s'" %
                             (self.target.__class__.__name__, name))

    def _run(self):
        for name, args, kwargs in self.calls:
            method = self._find_method(name)
            method(self.target, *args, **kwargs)


class LazySubReceiver(LazyReceiver):
    def __init__(self, name, target, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        super(LazySubReceiver, self).__init__(target)

    def __call__(self, target, *args, **kwargs):
        method = self._find_method(self.name)
        self.target = method(self.target, *self.args, **self.kwargs)
        self._run()


class LineJumpDrawFilter(LazyReceiver):
    def __init__(self, target, jump_radius):
        super(LineJumpDrawFilter, self).__init__(target)
        self.ytree = []
        self.x_cross = {}
        self.y_cross = {}
        self.forward = 'holizonal'
        self.jump_radius = jump_radius
        self.jump_shift = 0

    def set_options(self, **kwargs):
        if 'jump_forward' in kwargs:
            self.forward = kwargs['jump_forward']

        if 'jump_radius' in kwargs:
            self.jump_radius = kwargs['jump_radius']

        if 'jump_shift' in kwargs:
            self.jump_shift = kwargs['jump_shift']

    def _run(self):
        for name, args, kwargs in self.calls:
            if name == 'line' and kwargs.get('jump'):
                ((x1, y1), (x2, y2)) = args[0]
                if self.forward == 'holizonal' and y1 == y2:
                    self._holizonal_jumpline(x1, y1, x2, y2, **kwargs)
                    continue
                elif self.forward == 'vertical' and x1 == x2:
                    self._vertical_jumpline(x1, y1, x2, y2, **kwargs)
                    continue

            method = self._find_method(name)
            method(self.target, *args, **kwargs)

    def _holizonal_jumpline(self, x1, y1, x2, y2, **kwargs):
        y = y1
        if x2 < x1:
            x1, x2 = x2, x1

        for x in sorted(self.x_cross.get(y, [])):
            if x1 < x and x < x2:
                arckwargs = dict(kwargs)
                del arckwargs['jump']

                x += self.jump_shift
                r = self.jump_radius
                line = (XY(x1, y), XY(x - r, y))
                self.target.line(line, **kwargs)
                box = Box(x - r, y - r, x + r, y + r)
                self.target.arc(box, 180, 0, **arckwargs)
                x1 = x + r

        self.target.line((XY(x1, y), XY(x2, y)), **kwargs)

    def _vertical_jumpline(self, x1, y1, x2, y2, **kwargs):
        x = x1
        if y2 < y1:
            y1, y2 = y2, y1

        for y in sorted(self.y_cross.get(x, [])):
            if y1 < y and y < y2:
                arckwargs = dict(kwargs)
                del arckwargs['jump']

                y += self.jump_shift
                r = self.jump_radius
                line = (XY(x, y1), XY(x, y - r))
                self.target.line(line, **kwargs)
                box = Box(x - r, y - r, x + r, y + r)
                self.target.arc(box, 270, 90, **arckwargs)
                y1 = y + r

        self.target.line((XY(x, y1), XY(x, y2)), **kwargs)

    def line(self, xy, **kwargs):
        from bisect import insort
        for st, ed in zip(xy[:-1], xy[1:]):
            self.get_lazy_method("line")((st, ed), **kwargs)

            if 'jump' in kwargs and kwargs['jump'] is True:
                if st.y == ed.y:    # horizonal
                    insort(self.ytree, (st.y, 0, (st, ed)))
                elif st.x == ed.x:  # vertical
                    insort(self.ytree, (max(st.y, ed.y), -1, (st, ed)))
                    insort(self.ytree, (min(st.y, ed.y), +1, (st, ed)))

    def save(self, *args, **kwargs):
        # Search crosspoints
        from bisect import insort, bisect_left, bisect_right
        xtree = []
        for y, _, ((x1, y1), (x2, y2)) in self.ytree:
            if x2 < x1:
                x1, x2 = x2, x1
            if y2 < y1:
                y1, y2 = y2, y1

            if y == y1:
                insort(xtree, x1)

            if y == y2:
                del xtree[bisect_left(xtree, x1)]
                for x in xtree[bisect_right(xtree, x1):bisect_left(xtree, x2)]:
                    self.x_cross.setdefault(y, set()).add(x)
                    self.y_cross.setdefault(x, set()).add(y)

        self._run()
        return self.target.save(*args, **kwargs)



# -*- coding: utf-8 -*-
#  Copyright 2011 Takeshi KOMIYA
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from blockdiag import parser
from blockdiag.elements import Diagram, DiagramNode, NodeGroup, DiagramEdge
from blockdiag.plugins import fire_node_event
from blockdiag.utils import unquote, XY
from blockdiag.utils.compat import cmp_to_key


class DiagramTreeBuilder:
    def build(self, tree, config):
        self.config = config
        self.diagram = Diagram()
        self.instantiate(self.diagram, tree)
        for subgroup in self.diagram.traverse_groups():
            if len(subgroup.nodes) == 0:
                subgroup.group.nodes.remove(subgroup)

        self.bind_edges(self.diagram)
        self.fire_node_event('build_finished')
        return self.diagram

    def fire_node_event(self, event_type):
        for node in self.diagram.nodes:
            if node.drawable:
                fire_node_event(node, event_type)

    def is_related_group(self, group1, group2):
        if group1.is_parent(group2) or group2.is_parent(group1):
            return True
        else:
            return False

    def belong_to(self, node, group):
        if node.group and node.group.level > group.level:
            override = False
        else:
            override = True

        if node.group and node.group != group and override:
            if not self.is_related_group(node.group, group):
                msg = "could not belong to two groups: %s" % node.id
                raise RuntimeError(msg)

            old_group = node.group

            parent = group.parent(old_group.level + 1)
            if parent:
                if parent in old_group.nodes:
                    old_group.nodes.remove(parent)

                index = old_group.nodes.index(node)
                old_group.nodes.insert(index + 1, parent)

            old_group.nodes.remove(node)
            node.group = None

        if node.group is None:
            node.group = group

            if node not in group.nodes:
                group.nodes.append(node)

    def instantiate(self, group, tree):
        for stmt in tree.stmts:
            # Translate Node having group attribute to Group
            if isinstance(stmt, parser.Node):
                group_attr = [a for a in stmt.attrs if a.name == 'group']
                if group_attr:
                    group_id = group_attr[-1]
                    stmt.attrs.remove(group_id)

                    if group_id.value != group.id:
                        stmt = parser.Group(group_id.value, [stmt])

            # Instantiate statements
            if isinstance(stmt, parser.Node):
                node = DiagramNode.get(stmt.id)
                node.set_attributes(stmt.attrs)
                self.belong_to(node, group)

            elif isinstance(stmt, parser.Edge):
                from_nodes = [DiagramNode.get(n) for n in stmt.from_nodes]
                to_nodes = [DiagramNode.get(n) for n in stmt.to_nodes]

                for node in from_nodes + to_nodes:
                    self.belong_to(node, group)

                for node1 in from_nodes:
                    for node2 in to_nodes:
                        edge = DiagramEdge.get(node1, node2)
                        edge.set_dir(stmt.edge_type)
                        edge.set_attributes(stmt.attrs)

            elif isinstance(stmt, parser.Group):
                subgroup = NodeGroup.get(stmt.id)
                subgroup.level = group.level + 1
                self.belong_to(subgroup, group)
                self.instantiate(subgroup, stmt)

            elif isinstance(stmt, parser.Attr):
                group.set_attribute(stmt)

            elif isinstance(stmt, parser.Extension):
                if stmt.type == 'class':
                    name = unquote(stmt.name)
                    Diagram.classes[name] = stmt
                elif stmt.type == 'plugin':
                    self.diagram.set_plugin(stmt.name, stmt.attrs,
                                            config=self.config)

            elif isinstance(stmt, parser.Statements):
                self.instantiate(group, stmt)

        group.update_order()
        return group

    def bind_edges(self, group):
        for node in group.nodes:
            if isinstance(node, DiagramNode):
                group.edges += DiagramEdge.find(node)
            else:
                self.bind_edges(node)


class DiagramLayoutManager:
    def __init__(self, diagram):
        self.diagram = diagram

        self.circulars = []
        self.heightRefs = []
        self.coordinates = []

    def run(self):
        if isinstance(self.diagram, Diagram):
            for group in self.diagram.traverse_groups():
                self.__class__(group).run()

        self.edges = DiagramEdge.find_by_level(self.diagram.level)
        self.do_layout()
        self.diagram.fixiate()

        if self.diagram.orientation == 'portrait':
            self.rotate_diagram()

    def rotate_diagram(self):
        for node in self.diagram.traverse_nodes():
            node.xy = XY(node.xy.y, node.xy.x)
            node.colwidth, node.colheight = (node.colheight, node.colwidth)

            if isinstance(node, NodeGroup):
                if node.orientation == 'portrait':
                    node.orientation = 'landscape'
                else:
                    node.orientation = 'portrait'

        xy = (self.diagram.colheight, self.diagram.colwidth)
        self.diagram.colwidth, self.diagram.colheight = xy

    def do_layout(self):
        self.detect_circulars()

        self.set_node_xpos()
        self.adjust_node_order()

        height = 0
        for node in self.diagram.nodes:
            if node.xy.x == 0:
                self.set_node_ypos(node, height)
                height = max(xy.y for xy in self.coordinates) + 1

    def get_related_nodes(self, node, parent=False, child=False):
        uniq = {}
        for edge in self.edges:
            if edge.folded:
                continue

            if parent and edge.node2 == node:
                uniq[edge.node1] = 1
            elif child and edge.node1 == node:
                uniq[edge.node2] = 1

        related = []
        for uniq_node in uniq.keys():
            if uniq_node == node:
                pass
            elif uniq_node.group != node.group:
                pass
            else:
                related.append(uniq_node)

        related.sort(key=lambda x: x.order)
        return related

    def get_parent_nodes(self, node):
        return self.get_related_nodes(node, parent=True)

    def get_child_nodes(self, node):
        return self.get_related_nodes(node, child=True)

    def detect_circulars(self):
        for node in self.diagram.nodes:
            if not [x for x in self.circulars if node in x]:
                self.detect_circulars_sub(node, [node])

        # remove part of other circular
        for c1 in self.circulars[:]:
            for c2 in self.circulars:
                intersect = set(c1) & set(c2)

                if c1 != c2 and set(c1) == intersect:
                    if c1 in self.circulars:
                        self.circulars.remove(c1)
                    break

                if c1 != c2 and intersect:
                    if c1 in self.circulars:
                        self.circulars.remove(c1)
                    self.circulars.remove(c2)
                    self.circulars.append(c1 + c2)
                    break

    def detect_circulars_sub(self, node, parents):
        for child in self.get_child_nodes(node):
            if child in parents:
                i = parents.index(child)
                if parents[i:] not in self.circulars:
                    self.circulars.append(parents[i:])
            else:
                self.detect_circulars_sub(child, parents + [child])

    def is_circular_ref(self, node1, node2):
        for circular in self.circulars:
            if node1 in circular and node2 in circular:
                parents = []
                for node in circular:
                    for parent in self.get_parent_nodes(node):
                        if parent not in circular:
                            parents.append(parent)

                for parent in sorted(parents, key=lambda x: x.order):
                    children = self.get_child_nodes(parent)
                    if node1 in children and node2 in children:
                        if circular.index(node1) > circular.index(node2):
                            return True
                    elif node2 in children:
                        return True
                    elif node1 in children:
                        return False
                else:
                    if circular.index(node1) > circular.index(node2):
                        return True

        return False

    def set_node_xpos(self, depth=0):
        for node in self.diagram.nodes:
            if node.xy.x != depth:
                continue

            for child in self.get_child_nodes(node):
                if self.is_circular_ref(node, child):
                    pass
                elif node == child:
                    pass
                elif child.xy.x > node.xy.x + node.colwidth:
                    pass
                else:
                    child.xy = XY(node.xy.x + node.colwidth, 0)

        depther_node = [x for x in self.diagram.nodes if x.xy.x > depth]
        if len(depther_node) > 0:
            self.set_node_xpos(depth + 1)

    def adjust_node_order(self):
        for node in list(self.diagram.nodes):
            parents = self.get_parent_nodes(node)
            if len(set(parents)) > 1:
                for i in range(1, len(parents)):
                    node1 = parents[i - 1]
                    node2 = parents[i]

                    if node1.xy.x == node2.xy.x:
                        idx1 = self.diagram.nodes.index(node1)
                        idx2 = self.diagram.nodes.index(node2)

                        if idx1 < idx2:
                            self.diagram.nodes.remove(node2)
                            self.diagram.nodes.insert(idx1 + 1, node2)
                        else:
                            self.diagram.nodes.remove(node1)
                            self.diagram.nodes.insert(idx2 + 1, node1)

            children = self.get_child_nodes(node)
            if len(set(children)) > 1:
                for i in range(1, len(children)):
                    node1 = children[i - 1]
                    node2 = children[i]

                    idx1 = self.diagram.nodes.index(node1)
                    idx2 = self.diagram.nodes.index(node2)

                    if node1.xy.x == node2.xy.x:
                        if idx1 < idx2:
                            self.diagram.nodes.remove(node2)
                            self.diagram.nodes.insert(idx1 + 1, node2)
                        else:
                            self.diagram.nodes.remove(node1)
                            self.diagram.nodes.insert(idx2 + 1, node1)
                    elif self.is_circular_ref(node1, node2):
                        pass
                    else:
                        if node1.xy.x < node2.xy.x:
                            self.diagram.nodes.remove(node2)
                            self.diagram.nodes.insert(idx1 + 1, node2)
                        else:
                            self.diagram.nodes.remove(node1)
                            self.diagram.nodes.insert(idx2 + 1, node1)

            if isinstance(node, NodeGroup):
                children = self.get_child_nodes(node)
                if len(set(children)) > 1:
                    while True:
                        exchange = 0

                        for i in range(1, len(children)):
                            node1 = children[i - 1]
                            node2 = children[i]

                            idx1 = self.diagram.nodes.index(node1)
                            idx2 = self.diagram.nodes.index(node2)
                            ret = self.compare_child_node_order(node,
                                                                node1, node2)

                            if ret > 0 and idx1 < idx2:
                                self.diagram.nodes.remove(node1)
                                self.diagram.nodes.insert(idx2 + 1, node1)
                                exchange += 1

                        if exchange == 0:
                            break

        self.diagram.update_order()

    def compare_child_node_order(self, parent, node1, node2):
        def compare(x, y):
            x = x.duplicate()
            y = y.duplicate()
            while x.node1 == y.node1 and x.node1.group is not None:
                x.node1 = x.node1.group
                y.node1 = y.node1.group

            # cmp x.node1.order and y.node1.order
            if x.node1.order < y.node1.order:
                return -1
            elif x.node1.order == y.node1.order:
                return 0
            else:
                return 1

        edges = (DiagramEdge.find(parent, node1) +
                 DiagramEdge.find(parent, node2))
        edges.sort(key=cmp_to_key(compare))
        if len(edges) == 0:
            return 0
        elif edges[0].node2 == node2:
            return 1
        else:
            return -1

    def mark_xy(self, xy, width, height):
        for w in range(width):
            for h in range(height):
                self.coordinates.append(XY(xy.x + w, xy.y + h))

    def set_node_ypos(self, node, height=0):
        for x in range(node.colwidth):
            for y in range(node.colheight):
                xy = XY(node.xy.x + x, height + y)
                if xy in self.coordinates:
                    return False
        node.xy = XY(node.xy.x, height)
        self.mark_xy(node.xy, node.colwidth, node.colheight)

        def cmp(x, y):
            if x.xy.x < y.xy.y:
                return -1
            elif x.xy.x == y.xy.y:
                return 0
            else:
                return 1

        count = 0
        children = self.get_child_nodes(node)
        children.sort(key=cmp_to_key(cmp))

        grandchild = 0
        for child in children:
            if self.get_child_nodes(child):
                grandchild += 1

        prev_child = None
        for child in children:
            if child.id in self.heightRefs:
                pass
            elif node.xy.x >= child.xy.x:
                pass
            else:
                if isinstance(node, NodeGroup):
                    parent_height = self.get_parent_node_ypos(node, child)
                    if parent_height and parent_height > height:
                        height = parent_height

                if (prev_child and grandchild > 1 and
                   (not self.is_rhombus(prev_child, child))):
                    coord = [p.y for p in self.coordinates if p.x > child.xy.x]
                    if coord and max(coord) >= node.xy.y:
                        height = max(coord) + 1

                while True:
                    if self.set_node_ypos(child, height):
                        child.xy = XY(child.xy.x, height)
                        self.mark_xy(child.xy, child.colwidth, child.colheight)
                        self.heightRefs.append(child.id)

                        count += 1
                        break
                    else:
                        if count == 0:
                            return False

                        height += 1

                height += 1
                prev_child = child

        return True

    def is_rhombus(self, node1, node2):
        ret = False
        while True:
            if node1 == node2:
                ret = True
                break

            child1 = self.get_child_nodes(node1)
            child2 = self.get_child_nodes(node2)

            if len(child1) != 1 or len(child2) != 1:
                break
            elif node1.xy.x > child1[0].xy.x or node2.xy.x > child2[0].xy.x:
                break
            else:
                node1 = child1[0]
                node2 = child2[0]

        return ret

    def get_parent_node_ypos(self, parent, child):
        heights = []
        for e in DiagramEdge.find(parent, child):
            y = parent.xy.y

            node = e.node1
            while node != parent:
                y += node.xy.y
                node = node.group

            heights.append(y)

        if heights:
            return min(heights)
        else:
            return None


class EdgeLayoutManager(object):
    def __init__(self, diagram):
        self.diagram = diagram

    @property
    def groups(self):
        if self.diagram.separated:
            seq = self.diagram.nodes
        else:
            seq = self.diagram.traverse_groups(preorder=True)

        for group in seq:
            if not group.drawable:
                yield group

    @property
    def nodes(self):
        if self.diagram.separated:
            seq = self.diagram.nodes
        else:
            seq = self.diagram.traverse_nodes()

        for node in seq:
            if node.drawable:
                yield node

    @property
    def edges(self):
        for edge in (e for e in self.diagram.edges if e.style != 'none'):
            yield edge

        for group in self.groups:
            for edge in (e for e in group.edges if e.style != 'none'):
                yield edge

    def run(self):
        for edge in self.edges:
            _dir = edge.direction

            if edge.node1.group.orientation == 'landscape':
                if _dir == 'right':
                    r = range(edge.node1.xy.x + 1, edge.node2.xy.x)
                    for x in r:
                        xy = (x, edge.node1.xy.y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir == 'right-up':
                    r = range(edge.node1.xy.x + 1, edge.node2.xy.x)
                    for x in r:
                        xy = (x, edge.node1.xy.y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir == 'right-down':
                    if self.diagram.edge_layout == 'flowchart':
                        r = range(edge.node1.xy.y, edge.node2.xy.y)
                        for y in r:
                            xy = (edge.node1.xy.x, y + 1)
                            nodes = [x for x in self.nodes if x.xy == xy]
                            if len(nodes) > 0:
                                edge.skipped = 1

                    r = range(edge.node1.xy.x + 1, edge.node2.xy.x)
                    for x in r:
                        xy = (x, edge.node2.xy.y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir in ('left-down', 'down'):
                    r = range(edge.node1.xy.y + 1, edge.node2.xy.y)
                    for y in r:
                        xy = (edge.node1.xy.x, y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir == 'up':
                    r = range(edge.node2.xy.y + 1, edge.node1.xy.y)
                    for y in r:
                        xy = (edge.node1.xy.x, y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
            else:
                if _dir == 'right':
                    r = range(edge.node1.xy.x + 1, edge.node2.xy.x)
                    for x in r:
                        xy = (x, edge.node1.xy.y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir in ('left-down', 'down'):
                    r = range(edge.node1.xy.y + 1, edge.node2.xy.y)
                    for y in r:
                        xy = (edge.node1.xy.x, y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1
                elif _dir == 'right-down':
                    if self.diagram.edge_layout == 'flowchart':
                        r = range(edge.node1.xy.x, edge.node2.xy.x)
                        for x in r:
                            xy = (x + 1, edge.node1.xy.y)
                            nodes = [x for x in self.nodes if x.xy == xy]
                            if len(nodes) > 0:
                                edge.skipped = 1

                    r = range(edge.node1.xy.y + 1, edge.node2.xy.y)
                    for y in r:
                        xy = (edge.node2.xy.x, y)
                        nodes = [x for x in self.nodes if x.xy == xy]
                        if len(nodes) > 0:
                            edge.skipped = 1


class ScreenNodeBuilder:
    @classmethod
    def build(cls, tree, config=None, layout=True):
        DiagramNode.clear()
        DiagramEdge.clear()
        NodeGroup.clear()
        Diagram.clear()

        return cls(tree, config, layout).run()

    def __init__(self, tree, config, layout):
        self.diagram = DiagramTreeBuilder().build(tree, config)
        self.config = config
        self.layout = layout

    def run(self):
        if self.layout:
            DiagramLayoutManager(self.diagram).run()
            self.diagram.fixiate(True)

        EdgeLayoutManager(self.diagram).run()

        return self.diagram


class SeparateDiagramBuilder(ScreenNodeBuilder):
    @property
    def _groups(self):
        # Store nodes and edges of subgroups
        nodes = {self.diagram: self.diagram.nodes}
        edges = {self.diagram: self.diagram.edges}
        levels = {self.diagram: self.diagram.level}
        for group in self.diagram.traverse_groups():
            nodes[group] = group.nodes
            edges[group] = group.edges
            levels[group] = group.level

        groups = {}
        orders = {}
        for node in self.diagram.traverse_nodes():
            groups[node] = node.group
            orders[node] = node.order

        for group in self.diagram.traverse_groups():
            yield group

            # Restore nodes, groups and edges
            for g in nodes:
                g.nodes = nodes[g]
                g.edges = edges[g]
                g.level = levels[g]

            for n in groups:
                n.group = groups[n]
                n.order = orders[n]
                n.xy = XY(0, 0)
                n.colwidth = 1
                n.colheight = 1
                n.separated = False

            for edge in DiagramEdge.find_all():
                edge.skipped = False
                edge.crosspoints = []

        yield self.diagram

    def _filter_edges(self, edges, parent, level):
        filtered = {}
        for e in edges:
            if e.node1.group.is_parent(parent):
                if e.node1.group.level > level:
                    e = e.duplicate()
                    if isinstance(e.node1, NodeGroup):
                        e.node1 = e.node1.parent(level + 1)
                    else:
                        e.node1 = e.node1.group.parent(level + 1)
            else:
                continue

            if e.node2.group.is_parent(parent):
                if e.node2.group.level > level:
                    e = e.duplicate()
                    if isinstance(e.node2, NodeGroup):
                        e.node2 = e.node2.parent(level + 1)
                    else:
                        e.node2 = e.node2.group.parent(level + 1)
            else:
                continue

            filtered[(e.node1, e.node2)] = e

        return filtered.values()

    def run(self):
        for i, group in enumerate(self._groups):
            base = self.diagram.duplicate()
            base.level = group.level - 1

            # bind edges on base diagram (outer the group)
            edges = (DiagramEdge.find(None, group) +
                     DiagramEdge.find(group, None))
            base.edges = self._filter_edges(edges, self.diagram, group.level)

            # bind edges on target group (inner the group)
            subgroups = group.traverse_groups()
            edges = sum([g.edges for g in subgroups], group.edges)
            group.edges = []
            for e in self._filter_edges(edges, group, group.level):
                if isinstance(e.node1, NodeGroup) and e.node1 == e.node2:
                    pass
                else:
                    group.edges.append(e)

            # clear subgroups in the group
            for g in group.nodes:
                if isinstance(g, NodeGroup):
                    g.nodes = []
                    g.edges = []
                    g.separated = True

            # pick up nodes to base diagram
            nodes1 = [e.node1 for e in DiagramEdge.find(None, group)]
            nodes1.sort(key=lambda x: x.order)
            nodes2 = [e.node2 for e in DiagramEdge.find(group, None)]
            nodes2.sort(key=lambda x: x.order)

            nodes = nodes1 + [group] + nodes2
            for i, n in enumerate(nodes):
                n.order = i
                if n not in base.nodes:
                    base.nodes.append(n)
                    n.group = base

            if isinstance(group, Diagram):
                base = group

            DiagramLayoutManager(base).run()
            base.fixiate(True)
            EdgeLayoutManager(base).run()

            yield base
